from __future__ import annotations

import numpy as np

from matplotlib.patches import Rectangle
from typing import TYPE_CHECKING

from .rendering import Renderable
from .vehicle import Vehicle

import lanelet2
import lanelet2.geometry

if TYPE_CHECKING:
    import matplotlib.pyplot as plt
    from .road.road import Road


class EgoConfiguration(Renderable):
    def __init__(
        self,
        start_lanelet_id: int | None = None,
        start_s: float | None = None,
        start_t: float | None = None,
        start_x: float | None = None,
        start_y: float | None = None,
        start_lat: float | None = None,
        start_lon: float | None = None,
        z: float | None = None,
        target_lanelet_id: int | None = None,
        target_s: float | None = None,
        target_t: float | None = None,
        target_x: float | None = None,
        target_y: float | None = None,
        target_lat: float | None = None,
        target_lon: float | None = None,
        v0: float | None = None,
        vehicle_type_name: str = "medium",
        length: float | None = None,
        width: float | None = None,
        v_lon_max: float | None = None,
        a_lon_min: float | None = None,
        a_lon_max: float | None = None,
        controller: str | None = None,
    ) -> None:
        """
        start_lanelet_id: ID of initial lanelet the ego vehicle is starting on
        start_s: Initial longitudinal position relative to the given lanelet in m
        start_t: Initial lateral offset to the center line of the given lanelet in m
        start_x: Start x coordinate, used instead of start_lanelet_id/start_s/start_t if provided
        start_y: Start y coordinate, used instead of start_lanelet_id/start_s/start_t if provided
        start_lat: Start latitude coordinate, used instead of start_lanelet_id/start_s/start_t if provided
        start_lon: Start longitude coordinate, used instead of start_lanelet_id/start_s/start_t if provided
        z: Constant altitude in meters for OpenSCENARIO world positions
        target_lanelet_id: ID of the target lanelet, if None, start_lanelet_id is used
        target_s: Target longitudinal position in lanelet with target_lanelet_id
        target_t: Target lateral offset to the center line of the given lanelet in m
        target_x: Target x coordinate, used instead of target_s/target_t/target_lanelet_id if provided
        target_y: Target y coordinate, used instead of target_s/target_t/target_lanelet_id if provided
        target_lat: Target latitude coordinate, used instead of target_s/target_t/target_lanelet_id if provided
        target_lon: Target longitude coordinate, used instead of target_s/target_t/target_lanelet_id if provided
        v0: Initial speed in m/s
        vehicle_type_name: Type string of the vehicle, must be in Vehicle
        length: Length of the vehicle in m, if None, it is taken from the vehicle type
        width: Width of the vehicle in m, if None, it is taken from the vehicle type
        v_lon_max: Maximum longitudinal speed of the vehicle in m/s, if None, it is taken from the vehicle type
        a_lon_min: Minimum longitudinal acceleration of the vehicle in m/s^2, if None, it is taken from the vehicle type
        a_lon_max: Maximum longitudinal acceleration of the vehicle in m/s^2, if None, it is taken from the vehicle type
        controller: Controller for the ego vehicle
        """

        self._config = {
            "start_lanelet_id": start_lanelet_id,
            "start_s": start_s,
            "start_t": start_t,
            "start_x": start_x,
            "start_y": start_y,
            "start_lat": start_lat,
            "start_lon": start_lon,
            "z": z,
            "target_lanelet_id": target_lanelet_id,
            "target_s": target_s,
            "target_t": target_t,
            "target_x": target_x,
            "target_y": target_y,
            "target_lat": target_lat,
            "target_lon": target_lon,
            "v0": v0,
            "vehicle_type_name": vehicle_type_name,
            "length": length,
            "width": width,
            "v_lon_max": v_lon_max,
            "a_lon_min": a_lon_min,
            "a_lon_max": a_lon_max,
            "controller": controller,
        }

        self._start_position = {
            "lanelet_id": start_lanelet_id,
            "s": start_s,
            "t": start_t,
            "x": start_x,
            "y": start_y,
            "lat": start_lat,
            "lon": start_lon,
            "z": z,
        }

        self._target_position = {
            "lanelet_id": target_lanelet_id
            if target_lanelet_id is not None
            else start_lanelet_id,
            "s": target_s,
            "t": target_t,
            "x": target_x,
            "y": target_y,
            "lat": target_lat,
            "lon": target_lon,
        }

        self._v0 = v0
        self._vehicle_type_name = vehicle_type_name
        self._controller = controller

        if vehicle_type_name not in Vehicle.VEHICLE_TYPE_PARAMETERS:
            msg = f"Vehicle type {vehicle_type_name} not found. Available types: {Vehicle.VEHICLE_TYPE_PARAMETERS.keys()}"
            raise ValueError(msg)

        # Vehicle parameters
        if vehicle_type_name == "custom":
            if any(
                param is None
                for param in (length, width, v_lon_max, a_lon_min, a_lon_max)
            ):
                msg = "For custom vehicle, length, width, v_lon_max, a_lon_min, a_lon_max must be given."
                raise ValueError(msg)
            if length <= 0 or width <= 0:
                msg = "For custom vehicle, length and width must be greater than 0."
                raise ValueError(msg)
            if v_lon_max < 0:
                msg = "For custom vehicle, v_lon_max must be greater than 0."
                raise ValueError(msg)
            if a_lon_min > a_lon_max:
                msg = "For custom vehicle, a_lon_min must be less than a_lon_max."
                raise ValueError(msg)
            self._length = length
            self._width = width
            self._v_lon_max = v_lon_max
            self._a_lon_min = a_lon_min
            self._a_lon_max = a_lon_max
        else:
            vehicle_parameters = Vehicle.get_vehicle_parameters_of_vehicle_type(
                vehicle_type_name
            )
            self._length = vehicle_parameters.l
            self._width = vehicle_parameters.w
            self._v_lon_max = vehicle_parameters.longitudinal.v_max
            self._a_lon_min = -vehicle_parameters.longitudinal.a_max
            self._a_lon_max = vehicle_parameters.longitudinal.a_max

        self._traffic_rules = lanelet2.traffic_rules.create(
            lanelet2.traffic_rules.Locations.Germany,
            lanelet2.traffic_rules.Participants.Vehicle,
        )
        # Compiled values
        self._compiled = False
        self._start_heading = None
        self._route_length = None

    @property
    def config(self) -> dict:
        """
        Return the configuration as a dict.
        """
        return self._config

    @property
    def start_lanelet_id(self) -> int:
        """
        Lanelet ID of the start position.
        """
        self._check_is_compiled()
        return self._start_position["lanelet_id"]

    @property
    def start_s(self) -> float:
        """
        Start longitudinal position in lanelet coordinates.
        """
        self._check_is_compiled()
        return self._start_position["s"]

    @property
    def start_t(self) -> float:
        """
        Start lateral offset in lanelet coordinates.
        """
        self._check_is_compiled()
        return self._start_position["t"]

    @property
    def start_x(self) -> float:
        """
        x coordinate of the start position.
        """
        self._check_is_compiled()
        return self._start_position["x"]

    @property
    def start_y(self) -> float:
        """
        y coordinate of the start position.
        """
        self._check_is_compiled()
        return self._start_position["y"]

    @property
    def start_lat(self) -> float:
        """
        lat coordinate of the start position.
        """
        self._check_is_compiled()
        return self._start_position["lat"]

    @property
    def start_lon(self) -> float:
        """
        lon coordinate of the start position.
        """
        self._check_is_compiled()
        return self._start_position["lon"]

    @property
    def z(self) -> float:
        """
        Constant z coordinate of the ego vehicle.
        """
        self._check_is_compiled()
        return self._start_position["z"]

    @property
    def target_lanelet_id(self) -> int:
        """
        Lanelet ID of the target position.
        """
        self._check_is_compiled()
        return self._target_position["lanelet_id"]

    @property
    def target_s(self) -> float:
        """
        Target longitudinal position in lanelet coordinates.
        """
        self._check_is_compiled()
        return self._target_position["s"]

    @property
    def target_t(self) -> float:
        """
        Target lateral offset in lanelet coordinates.
        """
        self._check_is_compiled()
        return self._target_position["t"]

    @property
    def target_x(self) -> float:
        """
        x coordinate of the target position.
        """
        self._check_is_compiled()
        return self._target_position["x"]

    @property
    def target_y(self) -> float:
        """
        y coordinate of the target position.
        """
        self._check_is_compiled()
        return self._target_position["y"]

    @property
    def target_lat(self) -> float:
        """
        lat coordinate of the target position.
        """
        self._check_is_compiled()
        return self._target_position["lat"]

    @property
    def target_lon(self) -> float:
        """
        lon coordinate of the target position.
        """
        self._check_is_compiled()
        return self._target_position["lon"]

    @property
    def v0(self) -> float:
        """
        Initial speed in m/s.
        """
        return self._v0

    @property
    def length(self) -> float:
        """
        Vehicle length in meters.
        """
        return self._length

    @property
    def width(self) -> float:
        """
        Vehicle width in meters.
        """
        return self._width

    @property
    def v_lon_min(self) -> float:
        """
        Minimum longitudinal speed in m/s.
        """
        return 0

    @property
    def v_lon_max(self) -> float:
        """
        Maximum longitudinal speed in m/s.
        """
        return self._v_lon_max

    @property
    def a_lon_min(self) -> float:
        """
        Minimum longitudinal acceleration in m/s^2.
        """
        return self._a_lon_min

    @property
    def a_lon_max(self) -> float:
        """
        Maximum longitudinal acceleration in m/s^2.
        """
        return self._a_lon_max

    @property
    def start_heading(self) -> float:
        """
        Start heading in radians.
        """
        self._check_is_compiled()
        return self._start_heading

    @property
    def controller(self) -> str:
        """
        controller for ego vehicle.
        """
        return self._controller

    @property
    def route_length(self) -> float:
        """
        Length of the shortest route from start to target in meters.
        """
        self._check_is_compiled()
        return self._route_length

    def get_boundary_rect(self) -> tuple[float, float, float, float]:
        """
        Find the boundary rect.
        Returns (xmin, ymin, xmax, ymax).
        """
        xmin = min(self.start_x, self.target_x)
        xmax = max(self.start_x, self.target_x)
        ymin = min(self.start_y, self.target_y)
        ymax = max(self.start_y, self.target_y)
        return (xmin, ymin, xmax, ymax)

    @property
    def compiled(self) -> bool:
        """
        True if compiled values are available.
        """
        return self._compiled

    def _check_is_compiled(self) -> None:
        if not self._compiled:
            msg = "Please call .compile() before accessing this data."
            raise Exception(msg)

    def compile(self, road: Road) -> None:
        start_lanelet = road.resolve_position(self._start_position)
        target_lanelet = road.resolve_position(self._target_position)

        # Assumption: heading is along the lanelet
        # Moved along lanelet for 1s
        x1, y1 = road.from_frenet_to_cart(
            start_lanelet.centerline,
            self._start_position["s"] + 0.5,
            self._start_position["t"],
        )
        start_heading = np.arctan2(
            y1 - self._start_position["y"], x1 - self._start_position["x"]
        )

        # Calculate route length
        routing_graph = lanelet2.routing.RoutingGraph(
            road.lanelet_map, self._traffic_rules
        )
        route = routing_graph.getRoute(start_lanelet, target_lanelet)
        if route is None:
            msg = (
                "No valid route found. Please check the route "
                f"(start_lanelet_id: {self._start_position['lanelet_id']} -> target_lanelet_id: {self._target_position['lanelet_id']})."
            )
            raise ValueError(msg)

        target_lanelet_length = lanelet2.geometry.length2d(target_lanelet)
        route_length = route.length2d()
        route_length -= self._start_position["s"]
        route_length -= target_lanelet_length - min(
            self._target_position["s"], target_lanelet_length
        )

        if route_length < -1e-6:
            msg = (
                "Computed route length is negative. Please check that the target is "
                "ahead of the start position along the route."
            )
            raise ValueError(msg)

        # Compiled values
        self._start_heading = start_heading
        self._route_length = max(0.0, route_length)
        self._compiled = True

    def _plot_in_ax(self, ax: plt.Axes, *args, **kwargs) -> None:  # noqa: ARG002
        # Ego starting state

        def get_rotated_bbox_pts(
            center_x: float,
            center_y: float,
            heading: float,
            length: float,
            width: float,
        ) -> np.ndarray:
            T = np.array(  # noqa: N806
                [
                    [np.cos(heading), -np.sin(heading), 0, center_x],
                    [np.sin(heading), np.cos(heading), 0, center_y],
                    [0, 0, 1, 0],
                    [0, 0, 0, 1],
                ]
            )

            xmin = -length / 2
            xmax = length / 2
            ymin = -width / 2
            ymax = width / 2

            pt0 = np.dot(T, np.array([xmin, ymin, 0, 1]))[:2]
            pt1 = np.dot(T, np.array([xmax, ymin, 0, 1]))[:2]
            pt2 = np.dot(T, np.array([xmax, ymax, 0, 1]))[:2]
            pt3 = np.dot(T, np.array([xmin, ymax, 0, 1]))[:2]

            pts = np.array([pt0, pt1, pt2, pt3])
            return pts

        pts = get_rotated_bbox_pts(
            self.start_x, self.start_y, self._start_heading, self.length, self.width
        )

        pts_closed = np.vstack((pts, pts[0]))
        ax.plot(pts_closed[:, 0], pts_closed[:, 1], "b-", zorder=20, label="Ego start")

        rect_patch = Rectangle(
            (self.start_x - self.length / 2, self.start_y - self.width / 2),
            self.length,
            self.width,
            angle=np.rad2deg(self._start_heading),
            rotation_point="center",
            alpha=0.25,
            zorder=20,
            facecolor="b",
            edgecolor="b",
            label="Ego start",
        )
        ax.add_patch(rect_patch)

        # Ego target position
        ax.plot(
            self._target_position["x"],
            self._target_position["y"],
            "gx",
            markersize=5,
            zorder=20,
            label="Ego target",
        )

    def _format_ax(self, ax: plt.Axes, *args, **kwargs) -> None:  # noqa: ARG002
        ax.grid()
        ax.legend()

        ax.set(
            title="Ego configuration",
            xlabel="X position in m",
            ylabel="Y position in m",
        )

        ax.set_aspect("equal")



import pytest

from pathlib import Path

from simple_scenario.road import MappedRoad


class TestMappedRoad:
    DATA_DIR = Path(__file__).parent / "assets" / "synthetic-curve"
    OPENDRIVE_MAP = DATA_DIR / "synthetic-curve.xodr"
    LANELET2_MAP = DATA_DIR / "synthetic-curve.osm"

    RESULT_DIR = Path(__file__).parent / "results" / "test_mapped_road"
    RESULT_DIR.mkdir(exist_ok=True)

    @staticmethod
    def _create_result_dir(test_name: str) -> Path:
        result_dir = TestMappedRoad.RESULT_DIR / test_name
        result_dir.mkdir(exist_ok=True)
        return result_dir

    @staticmethod
    def _stress_test(test_name: str, road: MappedRoad, result_dir: Path) -> None:
        # Standard methods
        print(f"lat: {road.origin_lat}")
        print(f"lon: {road.origin_lon}")
        print(f"Saved llt2 at: {road.save_lanelet2_map(result_dir)}")
        print(f"Saved xodr at: {road.save_opendrive_map(result_dir)}")

        # Custom methods
        print(f"Original llt2 file: {road.lanelet2_map_file}")
        print(f"Original xodr file: {road.opendrive_map_file}")

        road.render(result_dir, plot_name=test_name)

    def test_mapped_road(self):
        test_name = "test_mapped_road_from_api"
        result_dir = self._create_result_dir(test_name)

        road = MappedRoad(self.OPENDRIVE_MAP, self.LANELET2_MAP)

        self._stress_test(test_name, road, result_dir)

    def test_mapped_road_wrong_file(self):
        invalid_map = "doesnotexist"

        with pytest.raises(
            FileNotFoundError, match=f"File {invalid_map} does not exist"
        ):
            MappedRoad(invalid_map, self.LANELET2_MAP)

        with pytest.raises(
            FileNotFoundError, match=f"File {invalid_map} does not exist"
        ):
            MappedRoad(self.OPENDRIVE_MAP, invalid_map)

        with pytest.raises(
            FileNotFoundError, match=f"File {invalid_map} does not exist"
        ):
            MappedRoad(invalid_map, invalid_map)


if __name__ == "__main__":
    tester = TestMappedRoad()
    tester.test_mapped_road()



from __future__ import annotations

import lanelet2
import numpy as np
import pyproj
import shutil

from lanelet2.core import LaneletMap, BasicPoint3d
from lanelet2.projection import UtmProjector
from lxml import etree
from pathlib import Path

from .road import Road


class MappedRoad(Road):
    def __init__(
        self, opendrive_map_file: str | Path, lanelet2_map_file: str | Path
    ) -> None:
        # Check inputs
        opendrive_map_file = Path(opendrive_map_file)
        lanelet2_map_file = Path(lanelet2_map_file)

        for file in (opendrive_map_file, lanelet2_map_file):
            if not file.exists():
                msg = f"File {file} does not exist"
                raise FileNotFoundError(msg)

        self._opendrive_map_file = opendrive_map_file
        self._lanelet2_map_file = lanelet2_map_file

        # Find xodr's proj
        self._xodr_proj_str = self._load_xodr_proj_str(self._opendrive_map_file)
        self._xodr_proj = pyproj.Proj(projparams=self._xodr_proj_str)
        # Extrat lat, lon origin
        self._origin_lon, self._origin_lat = self._xodr_proj(0, 0, inverse=True)
        # Find corresponding utm epsg
        utm_crs_list = pyproj.database.query_utm_crs_info(
            datum_name="WGS 84",
            area_of_interest=pyproj.aoi.AreaOfInterest(
                west_lon_degree=self._origin_lon,
                south_lat_degree=self._origin_lat,
                east_lon_degree=self._origin_lon,
                north_lat_degree=self._origin_lat,
            ),
        )
        if len(utm_crs_list) != 1:
            msg = "Cannot find exactly one UTM zone for the given OpenDrive origin"
            raise ValueError(msg)
        utm_epsg_code = int(utm_crs_list[0].code)
        self._utm_proj = pyproj.Proj(utm_epsg_code)

        # Load lanelet2 map
        self._llt_utm_projector = UtmProjector(
            lanelet2.io.Origin(self._origin_lat, self._origin_lon)
        )
        self._lanelet_map = lanelet2.io.load(
            str(self._lanelet2_map_file),
            self._llt_utm_projector,
        )

    def _load_xodr_proj_str(self, opendrive_map_file: str | Path) -> str:
        """
        Read origin lat lon from the opendrive map file's geoReference proj string.
        """

        # Find proj string
        tree = etree.parse(str(opendrive_map_file))
        proj_str = tree.getroot().find(".//geoReference").text

        if not proj_str:
            msg = "Could not find geoReference string in opendrive map"
            raise ValueError(msg)

        return proj_str

    def copy(self) -> MappedRoad:
        return MappedRoad(
            opendrive_map_file=self._opendrive_map_file,
            lanelet2_map_file=self._lanelet2_map_file,
        )

    @property
    def config(self) -> dict:
        return {
            "opendrive_map_file": str(self._opendrive_map_file),
            "lanelet2_map_file": str(self._lanelet2_map_file),
        }

    @property
    def opendrive_map_file(self) -> Path:
        return self._opendrive_map_file

    @property
    def lanelet2_map_file(self) -> Path:
        return self._lanelet2_map_file

    @property
    def origin_lat(self) -> float:
        return self._origin_lat

    @property
    def origin_lon(self) -> float:
        return self._origin_lon

    @property
    def lanelet_map(self) -> LaneletMap:
        return self._lanelet_map

    @property
    def llt_utm_projector(self) -> UtmProjector:
        return self._llt_utm_projector

    def from_llt_local_to_opendrive_local(
        self, x_llt2: float, y_llt2: float, heading_llt2: float | None = None
    ) -> tuple[float, float] | tuple[float, float, float]:
        """
        Transform a position from the lanelet maps local coordinate frame to the opendrive maps local coordinate frame.
        The lanelet2 map uses a transform from geographic coordinates (lat, lon) to a coordinate frame in UTM coordinates (relative to the origin).
        The opendrive map uses a transform given in the xodr file.

        Idea: Transform from llt2 local coordinates to lat, lon, then transform from lat, lon to xodr local coordinates.
        """

        # From llt2's local utm coordinates to lat, lon
        gps_point = self._llt_utm_projector.reverse(BasicPoint3d(x_llt2, y_llt2, 0))
        lat, lon = gps_point.lat, gps_point.lon

        # From lat, lon to xodr's local coordinates
        x_xodr, y_xodr = self._xodr_proj(lon, lat)

        if heading_llt2 is None:
            return x_xodr, y_xodr

        # Grid convergence
        grid_convergence_llt2 = np.deg2rad(
            self._utm_proj.get_factors(lon, lat).meridian_convergence
        )
        grid_convergence_xodr = np.deg2rad(
            self._xodr_proj.get_factors(lon, lat).meridian_convergence
        )
        heading_xodr = heading_llt2 - grid_convergence_llt2 + grid_convergence_xodr

        # Alternative: 2 point method (same results)
        # pt2x = x_llt2 + np.cos(heading_llt2)  # noqa: ERA001
        # pt2y = y_llt2 + np.sin(heading_llt2)  # noqa: ERA001
        # gps_point2 = self._llt_utm_projector.reverse(BasicPoint3d(pt2x, pt2y, 0))  # noqa: ERA001
        # heading_ = np.arctan2(pt2y - y_llt2, pt2x - x_llt2)  # noqa: ERA001
        # pt2x_xodr, pt2y_xodr = self._xodr_proj(gps_point2.lon, gps_point2.lat)  # noqa: ERA001
        # heading_xodr_ = np.arctan2(pt2y_xodr - y_xodr, pt2x_xodr - x_xodr)  # noqa: ERA001

        return x_xodr, y_xodr, heading_xodr

    def save_lanelet2_map(
        self, save_dir: str | Path, map_name: str | None = None
    ) -> Path:
        """
        Save the lanelet2 map to the specified directory.
        """
        save_dir = Path(save_dir)
        if not save_dir.is_dir():
            msg = "Save directory does not exist"
            raise FileNotFoundError(msg)

        if map_name is None:
            map_name = self._lanelet2_map_file.stem

        save_file = save_dir / (map_name + self._lanelet2_map_file.suffix)

        # Copy the original file to the save_dir
        shutil.copyfile(self._lanelet2_map_file, save_file)

        return save_file

    def save_opendrive_map(
        self, save_dir: str | Path, map_name: str | None = None
    ) -> Path:
        """
        Save the opendrive map to the specified directory.
        """
        save_dir = Path(save_dir)
        if not save_dir.is_dir():
            msg = "Save directory does not exist"
            raise FileNotFoundError(msg)

        if map_name is None:
            map_name = self._opendrive_map_file.stem

        save_file = save_dir / (map_name + self._opendrive_map_file.suffix)

        # Copy the original file to the save_dir
        shutil.copyfile(self._opendrive_map_file, save_file)

        return save_file



from __future__ import annotations

import json
import numpy as np
import pytest
import xml.etree.ElementTree as ET

from pathlib import Path

from simple_scenario import (
    Scenario,
    EgoConfiguration,
    Vehicle,
    CR_AVAILABLE,
)
from simple_scenario.rendering import create_scenario_plot_ax
from simple_scenario.road import SyntheticRoad, StraightSegment


class TestScenario:
    DATA_DIR = Path(__file__).parent / "assets"
    RESULT_DIR = Path(__file__).parent / "results" / "test_scenario"
    RESULT_DIR.mkdir(exist_ok=True)

    @staticmethod
    def _check_feasible(scenario: Scenario):
        if CR_AVAILABLE:
            assert scenario.is_feasible()
        else:
            with pytest.raises(ModuleNotFoundError):
                assert scenario.is_feasible()

    @staticmethod
    def _get_cr_interface(scenario: Scenario):
        if CR_AVAILABLE:
            scenario.get_cr_interface()
        else:
            with pytest.raises(ModuleNotFoundError):
                scenario.get_cr_interface()

    @staticmethod
    def _save_to_file(scenario: Scenario, result_dir: Path):
        if scenario._initialized_from_data:  # noqa: SLF001
            with pytest.raises(
                ValueError,
                match="Cannot save to config if the scenario has been initialized from data.",
            ):
                scenario.save(result_dir)
            with pytest.raises(
                ValueError,
                match="Cannot save to openx if the scenario has been initialized from data.",
            ):
                scenario.save(result_dir, mode="openx")
            with pytest.raises(
                ValueError,
                match="Cannot save to lanelet2 if the scenario has been initialized from data.",
            ):
                scenario.save(result_dir, mode="lanelet2")
        else:
            scenario.save(result_dir)
            scenario.save(result_dir, mode="openx")
            scenario.save(result_dir, mode="lanelet2")

        if CR_AVAILABLE:
            scenario.save(result_dir, mode="cr")
        else:
            with pytest.raises(ModuleNotFoundError):
                scenario.save(result_dir, mode="cr")

    @staticmethod
    def _get_world_positions(xosc_path: Path) -> list:
        root = ET.parse(xosc_path).getroot()  # noqa: S314
        return list(root.findall(".//WorldPosition"))

    @staticmethod
    def _minimal_openx_config(
        scenario_id: str,
        ego_z: float | None = None,
        vehicle_z: float | None = None,
    ) -> dict:
        config = {
            "scenario_id": scenario_id,
            "road": {
                "n_lanes": 2,
                "lane_width": 3.75,
                "segments": [{"length": 200, "heading": 0.0}],
                "speed_limit": 120,
                "x0": 0,
                "y0": 0,
            },
            "ego_configuration": {
                "start_lanelet_id": 1000,
                "start_s": 10,
                "start_t": 0,
                "target_s": 120,
                "target_t": 0,
                "v0": 10,
                "vehicle_type_name": "medium",
            },
            "vehicles": [
                {
                    "vehicle_id": 1,
                    "start_lanelet_id": 1000,
                    "start_s": 20,
                    "start_t": 0,
                    "v0": 8,
                    "vehicle_type_name": "medium",
                }
            ],
            "duration": 1.0,
            "dt": 0.2,
        }

        if ego_z is not None:
            config["ego_configuration"]["z"] = ego_z
        if vehicle_z is not None:
            config["vehicles"][0]["z"] = vehicle_z

        return config

    def test_scenario_creation(self):
        """
        Create a simple scenario in a pythonic way and save it to file (osm and json file!)
        """

        result_dir = self.RESULT_DIR / "test_scenario_creation"
        result_dir.mkdir(exist_ok=True)

        dt = 0.1
        duration = 10

        # Create road
        road = SyntheticRoad(3, 3.75, [StraightSegment(500, heading=0)])

        # Create ego configuration
        ego_configuration = EgoConfiguration(
            start_lanelet_id=1000,
            start_s=50,
            start_t=0,
            v0=27.78,
            target_s=450,
            target_t=0,
        )
        ego_configuration.compile(road)

        # Create vehicles

        # A vehicle in front of the ego vehicle
        vehicle0_thw0 = 3
        vehicle0_start_s = (
            ego_configuration.start_s + ego_configuration.v0 * vehicle0_thw0
        )
        vehicle0 = Vehicle(
            0,
            start_lanelet_id=ego_configuration.start_lanelet_id,
            start_s=vehicle0_start_s,
            start_t=0,
            v0=ego_configuration.v0,
        )

        # A vehicle in front of vehicle 0
        vehicle1_thw0 = 3
        vehicle1_start_s = vehicle0.start_s + vehicle0.v0 * vehicle1_thw0
        vehicle1 = Vehicle(
            1,
            start_lanelet_id=ego_configuration.start_lanelet_id,
            start_s=vehicle1_start_s,
            start_t=0,
            v0=vehicle0.v0,
        )

        # A vehicle on another lane
        vehicle2 = Vehicle(
            2,
            start_lanelet_id=1001,
            start_s=ego_configuration.start_s,
            start_t=0,
            v0=ego_configuration.v0,
        )

        vehicles = [vehicle0, vehicle1, vehicle2]

        # Create scenario
        scenario = Scenario("test", road, ego_configuration, vehicles, duration, dt=dt)

        # Render overview
        scenario.render(result_dir, "default")
        scenario.render(result_dir, "clean", clean=True)
        # Render timestep
        scenario.render(result_dir, timestep=50, plot_name_suffix="timestep_50")
        # Render GIF
        scenario.render_gif(result_dir, dpi=300)

        # Test ax mode
        with create_scenario_plot_ax(
            result_dir, f"scenario_{scenario.id}_axmode"
        ) as ax:
            scenario.render(ax)

        # Get cr interface
        self._get_cr_interface(scenario)

        # Save to file
        self._save_to_file(scenario, result_dir)

        # Check feasibility
        self._check_feasible(scenario)

    def test_scenario_creation_big_road(self):
        """
        Create a simple scenario in a pythonic way and save it to file (osm and json file!)
        """

        result_dir = self.RESULT_DIR / "test_scenario_creation_big_road"
        result_dir.mkdir(exist_ok=True)

        dt = 0.1
        duration = 10

        # Create road
        road = SyntheticRoad(5, 3.75, [StraightSegment(500, heading=0)])

        # Create ego configuration
        ego_configuration = EgoConfiguration(
            start_lanelet_id=1000,
            start_s=50,
            start_t=0,
            v0=27.78,
            target_s=450,
            target_t=0,
        )
        ego_configuration.compile(road)

        # Create vehicles

        # A vehicle in front of the ego vehicle
        vehicle0_thw0 = 3
        vehicle0_start_s = (
            ego_configuration.start_s + ego_configuration.v0 * vehicle0_thw0
        )
        vehicle0 = Vehicle(
            0,
            start_lanelet_id=ego_configuration.start_lanelet_id,
            start_s=vehicle0_start_s,
            start_t=0,
            v0=ego_configuration.v0,
        )

        # A vehicle in front of vehicle 0
        vehicle1_thw0 = 3
        vehicle1_start_s = vehicle0.start_s + vehicle0.v0 * vehicle1_thw0
        vehicle1 = Vehicle(
            1,
            start_lanelet_id=ego_configuration.start_lanelet_id,
            start_s=vehicle1_start_s,
            start_t=0,
            v0=vehicle0.v0,
        )

        # A vehicle on another lane
        vehicle2 = Vehicle(
            2,
            start_lanelet_id=1001,
            start_s=ego_configuration.start_s,
            start_t=0,
            v0=ego_configuration.v0,
        )

        vehicles = [vehicle0, vehicle1, vehicle2]

        # Create scenario
        scenario = Scenario(
            "test_big_road", road, ego_configuration, vehicles, duration, dt=dt
        )

        # Render overview
        scenario.render(result_dir, "default")
        scenario.render(result_dir, "clean", clean=True)
        # Render timestep
        scenario.render(result_dir, "timestep_50", timestep=50)
        # Render GIF
        scenario.render_gif(result_dir, dpi=300)

        # Test ax mode
        with create_scenario_plot_ax(
            result_dir, f"scenario_{scenario.id}_axmode"
        ) as ax:
            scenario.render(ax)

        # Get cr interface
        self._get_cr_interface(scenario)

        # Save to file
        self._save_to_file(scenario, result_dir)

        # Check feasibility
        self._check_feasible(scenario)

    def test_load_from_config(self):
        result_dir = self.RESULT_DIR / "test_load_from_config"
        result_dir.mkdir(exist_ok=True)

        config_file = self.DATA_DIR / "scenario_config.json"

        with config_file.open("r") as f:
            config = json.load(f)

        scenario = Scenario.from_config(config)

        scenario.render(result_dir, "from_config")

        # Save to file
        self._save_to_file(scenario, result_dir)

        # Check feasibility
        self._check_feasible(scenario)

    def test_openx_export_with_z(self):
        result_dir = self.RESULT_DIR / "test_openx_export_with_z"
        result_dir.mkdir(exist_ok=True)

        scenario = Scenario.from_config(
            self._minimal_openx_config("test_openx_z", ego_z=1.25, vehicle_z=2.5)
        )
        scenario.save(result_dir, mode="openx")

        assert np.isclose(scenario.ego_configuration.z, 1.25)
        assert np.isclose(scenario.vehicles[0].z, 2.5)

        xosc_path = result_dir / "test_openx_z.xosc"
        world_positions = self._get_world_positions(xosc_path)
        z_values = [
            float(position.attrib["z"])
            for position in world_positions
            if "z" in position.attrib
        ]

        assert len(z_values) == len(world_positions)
        assert any(np.isclose(z, 1.25) for z in z_values)
        assert any(np.isclose(z, 2.5) for z in z_values)

    def test_openx_export_without_z_keeps_previous_behavior(self):
        result_dir = self.RESULT_DIR / "test_openx_export_without_z"
        result_dir.mkdir(exist_ok=True)

        scenario = Scenario.from_config(self._minimal_openx_config("test_openx_no_z"))
        scenario.save(result_dir, mode="openx")

        xosc_path = result_dir / "test_openx_no_z.xosc"
        world_positions = self._get_world_positions(xosc_path)

        assert all("z" not in position.attrib for position in world_positions)

    def test_load_from_config_file(self):
        result_dir = self.RESULT_DIR / "test_load_from_config_file"
        result_dir.mkdir(exist_ok=True)

        config_file = self.DATA_DIR / "scenario_config.json"

        scenario = Scenario.from_config_file(config_file)

        scenario.render(result_dir, "from_config_file")

        # Get cr interface
        self._get_cr_interface(scenario)

        # Save to file
        self._save_to_file(scenario, result_dir)

        # Check feasibility
        self._check_feasible(scenario)

    def test_load_from_cr_xml(self):
        result_dir = self.RESULT_DIR / "test_load_from_config_file"
        result_dir.mkdir(exist_ok=True)

        xml = self.DATA_DIR / "ZAM_ValidationScenario+No+PaperScenario1-0_0_T-0.xml"

        if CR_AVAILABLE:
            scenario = Scenario.from_cr_xml(xml)

            scenario.render(result_dir, "from_cr_xml")

            # Save to file
            self._save_to_file(scenario, result_dir)

            # Check feasibility
            self._check_feasible(scenario)
        else:
            with pytest.raises(ModuleNotFoundError):
                scenario = Scenario.from_cr_xml(xml)

    def test_load_from_x_config(self):
        result_dir = self.RESULT_DIR / "test_load_from_x_config"
        result_dir.mkdir(exist_ok=True)

        config_file = self.DATA_DIR / "scenario_config.json"

        with config_file.open("r") as f:
            config = json.load(f)

        scenario = Scenario.from_x(config)

        scenario.render(result_dir, "from_x_config")

        # Get cr interface
        self._get_cr_interface(scenario)

        # Save to file
        self._save_to_file(scenario, result_dir)

        # Check feasibility
        self._check_feasible(scenario)

    def test_load_from_x_config_file(self):
        result_dir = self.RESULT_DIR / "test_load_from_x_config_file"
        result_dir.mkdir(exist_ok=True)

        config_file = self.DATA_DIR / "scenario_config.json"

        scenario = Scenario.from_x(config_file)

        scenario.render(result_dir, "from_x_config_file")

        # Get cr interface
        self._get_cr_interface(scenario)

        # Save to file
        self._save_to_file(scenario, result_dir)

        # Check feasibility
        self._check_feasible(scenario)

    def test_load_from_x_cr_xml(self):
        result_dir = self.RESULT_DIR / "test_load_from_x_cr_xml"
        result_dir.mkdir(exist_ok=True)

        xml = self.DATA_DIR / "ZAM_ValidationScenario+No+PaperScenario1-0_0_T-0.xml"

        if CR_AVAILABLE:
            scenario = Scenario.from_x(xml)

            scenario.render(result_dir, "from_x_cr_xml")

            # Get cr interface
            self._get_cr_interface(scenario)

            # Save to file
            self._save_to_file(scenario, result_dir)

            # Check feasibility
            self._check_feasible(scenario)
        else:
            with pytest.raises(ModuleNotFoundError):
                scenario = Scenario.from_x(xml)

    def test_load_from_x_object(self):
        result_dir = self.RESULT_DIR / "test_load_from_x_object"
        result_dir.mkdir(exist_ok=True)

        config_file = self.DATA_DIR / "scenario_config.json"

        scenario = Scenario.from_x(config_file)
        same_scenario = Scenario.from_x(scenario)

        same_scenario.render(result_dir, "test_load_from_x_object")

        # Get cr interface
        self._get_cr_interface(scenario)

        # Save to file
        self._save_to_file(scenario, result_dir)

        # Check feasibility
        self._check_feasible(scenario)

    def test_scenario_cr_interface(self):
        """
        Create a simple scenario in a pythonic way and save it to file (osm and json file!)
        """

        dt = 0.1
        duration = 10

        # Create road
        road = SyntheticRoad(3, 3.75, [StraightSegment(500, heading=0)])

        # Create ego configuration
        ego_configuration = EgoConfiguration(
            start_lanelet_id=1000,
            start_s=50,
            start_t=0,
            v0=27.78,
            target_s=450,
            target_t=0,
        )
        ego_configuration.compile(road)

        # Create vehicles

        # A vehicle in front of the ego vehicle
        vehicle0_thw0 = 3
        vehicle0_start_s = (
            ego_configuration.start_s + ego_configuration.v0 * vehicle0_thw0
        )
        vehicle0 = Vehicle(
            0,
            start_lanelet_id=ego_configuration.start_lanelet_id,
            start_s=vehicle0_start_s,
            start_t=0,
            v0=ego_configuration.v0,
        )

        # A vehicle in front of vehicle 0
        vehicle1_thw0 = 3
        vehicle1_start_s = vehicle0.start_s + vehicle0.v0 * vehicle1_thw0
        vehicle1 = Vehicle(
            1,
            start_lanelet_id=ego_configuration.start_lanelet_id,
            start_s=vehicle1_start_s,
            start_t=0,
            v0=vehicle0.v0,
        )

        # A vehicle on another lane
        vehicle2 = Vehicle(
            2,
            start_lanelet_id=1001,
            start_s=ego_configuration.start_s,
            start_t=0,
            v0=ego_configuration.v0,
        )

        vehicles = [vehicle0, vehicle1, vehicle2]

        # Create scenario
        scenario = Scenario("test", road, ego_configuration, vehicles, duration, dt=dt)

        # Get cr interface
        self._get_cr_interface(scenario)

        # Check feasibility
        self._check_feasible(scenario)

    def test_heading_calcuation_standstill(self):
        config_json = self.DATA_DIR / "scenario_standstill.json"
        scenario = Scenario.from_x(config_json)

        assert np.all(scenario.vehicles[0].heading != 0)
        self._get_cr_interface(scenario)

        self._check_feasible(scenario)


if __name__ == "__main__":
    tester = TestScenario()
    tester.test_scenario_creation()



import pytest

from pathlib import Path

from simple_scenario import (
    Scenario,
)


class TestScenarioOnMappedRoad:
    DATA_DIR = Path(__file__).parent / "assets" / "synthetic-curve"
    RESULT_DIR = Path(__file__).parent / "results" / "test_scenario_on_mapped_road"
    RESULT_DIR.mkdir(exist_ok=True)

    @staticmethod
    def _create_result_dir(test_name: str) -> Path:
        result_dir = TestScenarioOnMappedRoad.RESULT_DIR / test_name
        result_dir.mkdir(exist_ok=True)
        return result_dir

    @staticmethod
    def _stress_test(scenario: Scenario, result_dir: str) -> Path:
        scenario.save(result_dir, mode="openx")
        scenario.save(result_dir, mode="lanelet2")

        scenario.render(result_dir)
        with pytest.raises(
            NotImplementedError, match="Cannot create gif of mapped road."
        ):
            scenario.render_gif(result_dir)

    def test_lane_change(self):
        result_dir = self._create_result_dir("test_mapped_road_lane_change")
        scenario_config = self.DATA_DIR / "lane_change.json"

        scenario = Scenario.from_x(scenario_config)

        self._stress_test(scenario, result_dir)


if __name__ == "__main__":
    tester = TestScenarioOnMappedRoad()
    tester.test_lane_change()



from pathlib import Path

from scripts import create_scenario


class TestScripts:
    DATA_DIR = Path(__file__).parent / ".." / "scripts"

    def test_scripts_create_scenario(self):
        scenario_config = self.DATA_DIR / "simple_scenario.json"

        result_dir = create_scenario.main(
            scenario_config,
        )

        # Check that all files have been created
        assert (result_dir / f"{scenario_config.stem}.xosc").is_file()
        assert (result_dir / f"{scenario_config.stem}.xodr").is_file()
        assert (result_dir / f"{scenario_config.stem}.osm").is_file()
        assert (result_dir / f"{scenario_config.stem}.png").is_file()



from __future__ import annotations

import lanelet2
import numpy as np

from abc import abstractmethod
from typing import TYPE_CHECKING

from lanelet2.core import (
    getId,
    LaneletMap,
    LineString3d,
    Point3d,
    BasicPoint2d,
    GPSPoint,
)
from lanelet2.geometry import findNearest

if TYPE_CHECKING:
    import matplotlib.pyplot as plt
    from pathlib import Path

from ..rendering import Renderable


class Road(Renderable):
    @abstractmethod
    def copy(self) -> Road:
        """
        Create a copy of the road object.
        """
        ...

    @property
    @abstractmethod
    def config(self) -> dict:
        """
        Return the config of the road object as a dict.
        """
        ...

    @property
    @abstractmethod
    def lanelet_map(self) -> LaneletMap: ...

    @property
    @abstractmethod
    def origin_lat(self) -> Path:
        """
        Latitude of the origin of the map.
        """
        ...

    @property
    @abstractmethod
    def origin_lon(self) -> Path:
        """
        Longitude of the origin of the map.
        """
        ...

    def get_boundary_rect(self) -> tuple[float, float, float, float]:
        """
        Find the boundary rect of the road.
        Returns (xmin, ymin, xmax, ymax)
        """

        xmin = None
        ymin = None
        xmax = None
        ymax = None

        for lanelet in self.lanelet_map.laneletLayer:
            for line in [lanelet.leftBound, lanelet.rightBound]:
                for pt in line:
                    if xmin is None or pt.x < xmin:
                        xmin = pt.x
                    if ymin is None or pt.y < ymin:
                        ymin = pt.y
                    if xmax is None or pt.x > xmax:
                        xmax = pt.x
                    if ymax is None or pt.y > ymax:
                        ymax = pt.y

        return (xmin, ymin, xmax, ymax)

    @abstractmethod
    def save_lanelet2_map(self, save_dir: str | Path, map_name: str) -> Path:
        """
        Save the lanelet2 map to the specified directory.
        """
        ...

    @abstractmethod
    def save_opendrive_map(self, save_dir: str | Path, map_name: str) -> Path:
        """
        Save the opendrive map to the specified directory.
        """
        ...

    @staticmethod
    def linestring2array(linestring: LineString3d) -> np.array:
        """
        Convert a linestring to an array of x, y points.
        """
        return np.array([[pt.x, pt.y] for pt in linestring])

    @staticmethod
    def array2linestring(array: np.array) -> LineString3d:
        """
        Convert an array of x, y points to a linestring.
        """
        return LineString3d(
            getId(), [Point3d(getId(), pt[0], pt[1], 0.0) for pt in array]
        )

    @staticmethod
    def from_frenet_to_cart(
        linestring: LineString3d, s: np.ndarray, t: np.ndarray
    ) -> tuple[np.ndarray, np.ndarray]:
        """
        Convert frenet coordinates (s, t) along a linestring to cartesian coordinates (x, y).
        """

        def check_array(array_or_number: np.ndarray | float) -> np.ndarray:
            if isinstance(array_or_number, (float, int)):
                array_or_number = np.array([array_or_number]).astype(float)
            return array_or_number

        s = check_array(s)
        t = check_array(t)

        if s.shape != t.shape:
            raise Exception

        linestring2d = lanelet2.geometry.to2D(linestring)

        x = np.zeros_like(s)
        y = np.zeros_like(t)

        for i in range(s.shape[0]):
            arc = lanelet2.geometry.ArcCoordinates()
            arc.length = s[i]
            arc.distance = t[i]

            cart = lanelet2.geometry.fromArcCoordinates(linestring2d, arc)

            x[i] = cart.x
            y[i] = cart.y

        if x.shape[0] == 1:
            x = x[0]
            y = y[0]

        return x, y

    @staticmethod
    def from_cart_to_frenet(
        linestring: LineString3d, x: np.ndarray, y: np.ndarray
    ) -> tuple[np.ndarray, np.ndarray]:
        """
        Convert cartesian coordinates (x, y) to frenet coordinates (s, t) along a linestring.
        """

        def check_array(array_or_number: np.ndarray | float) -> np.ndarray:
            if isinstance(array_or_number, (float, int)):
                array_or_number = np.array([array_or_number]).astype(float)
            return array_or_number

        x = check_array(x)
        y = check_array(y)

        if x.shape != y.shape:
            raise Exception

        linestring2d = lanelet2.geometry.to2D(linestring)

        s = np.zeros_like(x)
        t = np.zeros_like(y)

        for i in range(x.shape[0]):
            arc = lanelet2.geometry.toArcCoordinates(
                linestring2d, BasicPoint2d(x[0], y[0])
            )

            s[i] = arc.length
            t[i] = arc.distance

        if s.shape[0] == 1:
            s = s[0]
            t = t[0]

        return s, t

    def find_lanelet_id_by_position(self, x: float, y: float) -> int:
        return self.find_lanelet_id_by_position_on_lanelet_map(self.lanelet_map, x, y)

    @staticmethod
    def find_lanelet_id_by_position_on_lanelet_map(
        lanelet_map: LaneletMap, x: float, y: float
    ) -> int:
        """
        Find ID of lanelet that x, y is inside.
        Return None if x,y is not inside any lanelet or in more than one lanelets.
        """
        candidates = findNearest(lanelet_map.laneletLayer, BasicPoint2d(x, y), 1)

        if len(candidates) == 0 or candidates[0][0] != 0:
            return None

        llt_id = candidates[0][1].id

        return llt_id

    @abstractmethod
    def from_llt_local_to_opendrive_local(
        self, x_llt2: float, y_llt2: float, heading_llt2: float | None = None
    ) -> tuple[float, float] | tuple[float, float, float]: ...

    def _plot_in_ax(self, ax: plt.Axes) -> None:
        """
        Plot the lanelet map in the given ax.
        """

        plotted_lanelines = set()

        for lanelet in self.lanelet_map.laneletLayer:
            leftbound = self.linestring2array(lanelet.leftBound)
            rightbound = self.linestring2array(lanelet.rightBound)

            clr = "k"

            # Find shape (dashed/solid)
            left_shape = "-"
            if (
                "subtype" in lanelet.leftBound.attributes
                and lanelet.leftBound.attributes["subtype"] == "dashed"
            ):
                left_shape = "--"

            right_shape = "-"
            if (
                "subtype" in lanelet.rightBound.attributes
                and lanelet.rightBound.attributes["subtype"] == "dashed"
            ):
                right_shape = "--"

            if lanelet.leftBound.id not in plotted_lanelines:
                ax.plot(leftbound[0, 0], leftbound[0, 1], "x", color=clr, zorder=10)
                ax.plot(
                    leftbound[:, 0],
                    leftbound[:, 1],
                    left_shape,
                    color=clr,
                    zorder=10,
                )
                ax.plot(leftbound[0, 0], leftbound[0, 1], "x", color=clr, zorder=10)
            if lanelet.rightBound.id not in plotted_lanelines:
                ax.plot(rightbound[0, 0], rightbound[0, 1], "x", color=clr, zorder=10)
                ax.plot(
                    rightbound[:, 0],
                    rightbound[:, 1],
                    right_shape,
                    color=clr,
                    zorder=10,
                )
                ax.plot(rightbound[0, 0], rightbound[0, 1], "x", color=clr, zorder=10)

            plotted_lanelines.add(lanelet.leftBound.id)
            plotted_lanelines.add(lanelet.rightBound.id)

    def _format_ax(self, ax: plt.Axes) -> None:
        """
        Format the ax after plotting.
        """

        ax.set(title="Road", xlabel="X position in m", ylabel="Y position in m")

        # Set axis limits
        margin = 10
        xmin, ymin, xmax, ymax = self.get_boundary_rect()
        ax.set_xlim(xmin - margin, xmax + margin)
        ax.set_ylim(ymin - margin, ymax + margin)

        ax.set_aspect("equal")

    def resolve_position(
        self, position: dict[str, float | int | None]
    ) -> lanelet2.core.Lanelet:
        if position["lat"] is not None and position["lon"] is not None:
            if self.llt_utm_projector is None:
                msg = "Start lat/lon requires a lanelet2 projector."
                raise ValueError(msg)
            z = position.get("z")
            gps = GPSPoint(position["lat"], position["lon"], 0.0 if z is None else z)
            utm = self.llt_utm_projector.forward(gps)
            position["x"], position["y"] = utm.x, utm.y

        if position["x"] is not None and position["y"] is not None:
            position["lanelet_id"] = self.find_lanelet_id_by_position(
                position["x"], position["y"]
            )
            if position["lanelet_id"] is None:
                msg = "Start position is not located in any lanelet."
                raise ValueError(msg)
            lanelet = self.lanelet_map.laneletLayer[position["lanelet_id"]]
            position["s"], position["t"] = self.from_cart_to_frenet(
                lanelet.centerline, position["x"], position["y"]
            )

        if (
            position["lanelet_id"] is not None
            and position["s"] is not None
            and position["t"] is not None
            and (position["x"] is None or position["y"] is None)
        ):
            lanelet = self.lanelet_map.laneletLayer[position["lanelet_id"]]
            position["x"], position["y"] = self.from_frenet_to_cart(
                lanelet.centerline, position["s"], position["t"]
            )

        # Check if lanelet is valid
        if position["lanelet_id"] is None:
            msg = "lanelet ID could not be determined."
            raise ValueError(msg)

        return self.lanelet_map.laneletLayer[position["lanelet_id"]]



from pathlib import Path

from simple_scenario.cli import create


class TestCLI:
    DATA_DIR = Path(__file__).parent / ".." / "demo"

    RESULT_DIR = Path(__file__).parent / "results" / "test_cli"
    RESULT_DIR.mkdir(exist_ok=True)

    def test_cli_demo_default(self):
        result_dir = self.RESULT_DIR / "test_cli_demo_default"
        result_dir.mkdir(exist_ok=True)

        scenario_config = self.DATA_DIR / "cutout_example.json"

        create(
            scenario_config,
            result_dir,
        )

        # Check that all files have been created
        openx_dir = result_dir / f"{scenario_config.stem}_openx"
        assert openx_dir.is_dir()
        assert (openx_dir / f"{scenario_config.stem}.xosc").is_file()
        assert (openx_dir / f"{scenario_config.stem}.xodr").is_file()

    def test_cli_demo_with_lanelet2(self):
        result_dir = self.RESULT_DIR / "test_cli_demo_with_lanelet2"
        result_dir.mkdir(exist_ok=True)

        scenario_config = self.DATA_DIR / "cutout_example.json"

        create(
            scenario_config,
            result_dir,
            lanelet2=True,
        )

        # Check that all files have been created
        openx_dir = result_dir / f"{scenario_config.stem}_openx"
        assert openx_dir.is_dir()
        assert (openx_dir / f"{scenario_config.stem}.xosc").is_file()
        assert (openx_dir / f"{scenario_config.stem}.xodr").is_file()
        assert (openx_dir / f"{scenario_config.stem}.osm").is_file()



from pathlib import Path

import numpy as np

from simple_scenario import EgoConfiguration
from simple_scenario.rendering import create_scenario_plot_ax
from simple_scenario.road import SyntheticRoad, StraightSegment


class TestEgoConfiguration:
    RESULT_DIR = Path(__file__).parent / "results" / "test_ego_configuration"
    RESULT_DIR.mkdir(exist_ok=True)

    def test_ego_configuration(self):
        result_dir = self.RESULT_DIR / "test_ego_configuration"
        result_dir.mkdir(exist_ok=True)

        road = SyntheticRoad(3, 3.75, [StraightSegment(200)])

        ego_configuration = EgoConfiguration(
            start_lanelet_id=1000,
            start_s=50,
            start_t=0,
            v0=27.78,
            target_s=150,
            target_t=0,
        )

        ego_configuration.compile(road)

        ego_configuration.render(result_dir, "ego_configuration")
        ego_configuration.render(result_dir, "ego_configuration_clean", clean=True)

        with create_scenario_plot_ax(result_dir, "ego_configration_on_road") as ax:
            road.render(ax)
            ego_configuration.render(ax)

    def test_ego_configuration_target_xy(self):
        result_dir = self.RESULT_DIR / "test_ego_configuration_target_xy"
        result_dir.mkdir(exist_ok=True)

        road = SyntheticRoad(3, 3.75, [StraightSegment(200)])

        target_lanelet_id = 1000
        target_s = 150
        target_t = 0.0
        target_x, target_y = road.from_frenet_to_cart(
            road.lanelet_map.laneletLayer[target_lanelet_id].centerline,
            target_s,
            target_t,
        )

        ego_configuration = EgoConfiguration(
            start_lanelet_id=target_lanelet_id,
            start_s=50,
            start_t=0,
            v0=27.78,
            target_s=None,
            target_t=None,
            target_lanelet_id=None,
            target_x=target_x,
            target_y=target_y,
        )

        ego_configuration.compile(road)

        assert ego_configuration.target_lanelet_id == target_lanelet_id
        assert np.isclose(ego_configuration.target_s, target_s)
        assert np.isclose(ego_configuration.target_t, target_t)
        assert np.isclose(ego_configuration.route_length, target_s - 50)

        with create_scenario_plot_ax(
            result_dir, "ego_configration_on_road_clean", clean=True
        ) as ax:
            road.render(ax)
            ego_configuration.render(ax)


if __name__ == "__main__":
    tester = TestEgoConfiguration()
    tester.test_ego_configuration()



from __future__ import annotations

import lanelet2
import lanelet2.geometry
import numpy as np

from matplotlib.patches import Rectangle
from typing import TYPE_CHECKING
from vehiclemodels.parameters_vehicle1 import parameters_vehicle1
from vehiclemodels.parameters_vehicle2 import parameters_vehicle2
from vehiclemodels.parameters_vehicle3 import parameters_vehicle3

from .road.road import Road
from .rendering import Renderable

if TYPE_CHECKING:
    import matplotlib.pyplot as plt
    from omegaconf import DictConfig
    from lanelet2.core import LaneletMap


class Vehicle(Renderable):
    VEHICLE_TYPE_PARAMETERS = {
        "small": parameters_vehicle1,  # VehicleType.FORD_ESCORT,
        "medium": parameters_vehicle2,  # VehicleType.BMW_320i,
        "van": parameters_vehicle3,  # VehicleType.VW_VANAGON
        # "semi_trailer": 4  # Currently not available
        "custom": None,  # Take width and length from parameters
    }

    ACCELERATION_PROFILES = ["constant"]

    LC_TYPES = ("polynomial", "vy")

    def __init__(
        self,
        vehicle_id: int,
        *,
        start_lanelet_id: int | None = None,
        start_s: float | None = None,
        start_t: float | None = None,
        start_x: float | None = None,
        start_y: float | None = None,
        start_lat: float | None = None,
        start_lon: float | None = None,
        z: float | None = None,
        target_lanelet_id: int | None = None,
        target_s: float | None = None,
        target_t: float | None = None,
        target_x: float | None = None,
        target_y: float | None = None,
        target_lat: float | None = None,
        target_lon: float | None = None,
        v0: float | None = None,
        a0: float = 0,
        a_delay: float = 0,
        a_profile: str = "constant",
        lc_direction: int = 0,
        lc_delay: float = 0,
        lc_duration: float = 3,
        lc_type: str = "polynomial",
        lc_vy: float = 0,
        inverse_driving_direction: bool = False,
        vehicle_type_name: str = "medium",
        length: float | None = None,
        width: float | None = None,
        from_data: bool = False,
        depends_on_ego: bool = False,
        duration: float | None = None,
    ) -> None:
        """
        vehicle_id: Unique ID of the vehicle
        start_lanelet_id: ID of initial lanelet the vehicle is starting on
        start_s: Initial longitudinal position relative to the given lanelet in m
        start_t: Initial lateral offset to the center line of the given lanelet in m
        start_x: Start x coordinate, used instead of start_lanelet_id/start_s/start_t if provided
        start_y: Start y coordinate, used instead of start_lanelet_id/start_s/start_t if provided
        start_lat: Start latitude coordinate, used instead of start_lanelet_id/start_s/start_t if provided
        start_lon: Start longitude coordinate, used instead of start_lanelet_id/start_s/start_t if provided
        z: Constant altitude in meters for OpenSCENARIO world positions
        target_lanelet_id: ID of the target lanelet, if None, start_lanelet_id is used
        target_s: Target longitudinal position in lanelet with target_lanelet_id
        target_t: Target lateral offset to the center line of the given lanelet in m
        target_x: Target x coordinate, used instead of target_s/target_t/target_lanelet_id if provided
        target_y: Target y coordinate, used instead of target_s/target_t/target_lanelet_id if provided
        target_lat: Target latitude coordinate, used instead of target_s/target_t/target_lanelet_id if provided
        target_lon: Target longitude coordinate, used instead of target_s/target_t/target_lanelet_id if provided
        v0: Initial speed in m/s
        a0: Acceleration in m/s^2
        a_delay: Time to wait before applying acceleration
        a_profile: How the acceleration changes over time
        lc_direction: Direction of lane change. 0: no lane change, 1: left, -1: right
        lc_delay: Time to wait before lane change
        lc_duration: Time the lane change takes to be completed
        lc_type: How the lc is done
        lc_vy: Only used if lc_type is "vy"
        inverse_driving_direction: If True, the vehicle is driving in the opposite direction
        vehicle_type_name: Type of the vehicle
        length: Length of the vehicle, only used if vehicle_type_name is "custom"
        width: Width of the vehicle, only used if vehicle_type_name is "custom"
        from_data: If True, the vehicle is created from data and no compilation is needed
        depends_on_ego: If True, the vehicle's behavior depends on the ego vehicle movement
        duration: Optional duration for this vehicle's trajectory in seconds
        """

        if a_profile not in self.ACCELERATION_PROFILES:
            msg = f"Unsupported acceleration profile specifed: {a_profile}. Choose from: {self.ACCELERATION_PROFILES}."
            raise AttributeError(msg)

        if lc_direction not in (-1, 0, 1):
            msg = "lc_direction can be -1, 0, or 1"
            raise ValueError(msg)

        if lc_type not in self.LC_TYPES:
            msg = f"Unsupported lc type specifed: {lc_type}. Choose from: {self.LC_TYPES}."
            raise AttributeError(msg)

        if vehicle_type_name not in self.VEHICLE_TYPE_PARAMETERS:
            msg = "Vehicle type is not available"
            raise ValueError(msg)

        if vehicle_type_name == "custom" and length is None and width is None:
            msg = "If vehicle type is 'custom', parameters 'length' and 'width' must be not None"
            raise ValueError(msg)

        if vehicle_type_name != "custom" and length is not None and width is not None:
            msg = "If vehicle type is not 'custom', parameters 'length' and 'width' must be None"
            raise ValueError(msg)

        self._config = {
            "vehicle_id": vehicle_id,
            "start_lanelet_id": start_lanelet_id,
            "start_s": start_s,
            "start_t": start_t,
            "start_x": start_x,
            "start_y": start_y,
            "start_lat": start_lat,
            "start_lon": start_lon,
            "z": z,
            "target_lanelet_id": target_lanelet_id,
            "target_s": target_s,
            "target_t": target_t,
            "target_x": target_x,
            "target_y": target_y,
            "target_lat": target_lat,
            "target_lon": target_lon,
            "v0": v0,
            "a0": a0,
            "a_delay": a_delay,
            "a_profile": a_profile,
            "lc_direction": lc_direction,
            "lc_delay": lc_delay,
            "lc_duration": lc_duration,
            "lc_type": lc_type,
            "lc_vy": lc_vy,
            "inverse_driving_direction": inverse_driving_direction,
            "vehicle_type_name": vehicle_type_name,
            "length": length,
            "width": width,
            "from_data": from_data,
            "depends_on_ego": depends_on_ego,
            "duration": duration,
        }

        self._start_position = {
            "lanelet_id": start_lanelet_id,
            "s": start_s,
            "t": start_t,
            "x": start_x,
            "y": start_y,
            "lat": start_lat,
            "lon": start_lon,
            "z": z,
        }

        self._target_position = {
            "lanelet_id": target_lanelet_id
            if target_lanelet_id is not None
            else start_lanelet_id,
            "s": target_s,
            "t": target_t,
            "x": target_x,
            "y": target_y,
            "lat": target_lat,
            "lon": target_lon,
        }

        self._vehicle_id = vehicle_id
        self._v0 = v0
        self._a0 = a0
        self._a_delay = a_delay
        self._a_profile = a_profile
        self._lc_direction = lc_direction
        self._lc_delay = lc_delay
        self._lc_duration = lc_duration
        self._lc_type = lc_type
        self._lc_vy = lc_vy
        self._inverse_driving_direction = inverse_driving_direction
        self._vehicle_type_name = vehicle_type_name
        self._vehicle_parameters = None
        if self._vehicle_type_name == "custom":
            self._length = length
            self._width = width
        else:
            vehicle_parameters = self.get_vehicle_parameters_of_vehicle_type(
                vehicle_type_name
            )
            self._length = vehicle_parameters.l
            self._width = vehicle_parameters.w
        self._depends_on_ego = depends_on_ego
        self._duration = duration

        self._traffic_rules = lanelet2.traffic_rules.create(
            lanelet2.traffic_rules.Locations.Germany,
            lanelet2.traffic_rules.Participants.Vehicle,
        )

        # Compiled values
        self._compiled = False
        self._x = None
        self._y = None
        self._heading = None
        self._v = None
        self._a = None

        # If the vehicle is created directly from data, this is True
        self._initialized_from_data = from_data

    @classmethod
    def get_vehicle_parameters_of_vehicle_type(cls, vehicle_type_name: str) -> dict:
        if vehicle_type_name not in cls.VEHICLE_TYPE_PARAMETERS:
            msg = "Vehicle type is not available"
            raise ValueError(msg)

        if vehicle_type_name == "custom":
            msg = "If vehicle type is 'custom', parameters must be selected manually."
            raise ValueError(msg)

        vehicle_parameters = cls.VEHICLE_TYPE_PARAMETERS[vehicle_type_name]()
        return vehicle_parameters

    @classmethod
    def from_data(
        cls,
        vehicle_id: int,
        x: np.ndarray,
        y: np.ndarray,
        heading: np.ndarray,
        v: np.ndarray,
        a: np.ndarray,
        vehicle_type_name: str = "medium",
        length: float | None = None,
        width: float | None = None,
        depends_on_ego: bool = False,
        duration: float | None = None,
    ) -> Vehicle:
        if x.shape[0] == 0:
            msg = "All data arrays must have some values."
            raise ValueError(msg)

        if len(x.shape) != 1:
            msg = "All data arrays must be of shape (X,)."
            raise ValueError(msg)

        if not (x.shape == y.shape == heading.shape == v.shape == a.shape):
            msg = "All data arrays must have same shape"
            raise ValueError(msg)

        vehicle = Vehicle(
            vehicle_id=vehicle_id,
            v0=v[0],
            a0=a[0],
            vehicle_type_name=vehicle_type_name,
            length=length,
            width=width,
            from_data=True,
            depends_on_ego=depends_on_ego,
            duration=duration,
        )

        vehicle.set_data(x, y, heading, v, a)

        return vehicle

    @property
    def config(self) -> dict:
        return self._config

    @property
    def id(self) -> int:
        return self._vehicle_id

    @property
    def vehicle_parameters(self) -> DictConfig:
        if self._vehicle_type_name == "custom":
            msg = (
                "If vehicle_type_name is 'custom', vehicle_parameters are not available"
            )
            raise ValueError(msg)
        return self._vehicle_parameters

    @property
    def start_lanelet_id(self) -> int:
        if self._initialized_from_data:
            msg = "Not possible to access start_lanelet_id when initialized from data."
            raise Exception(msg)
        return self._start_position["lanelet_id"]

    @property
    def start_s(self) -> float:
        if self._initialized_from_data:
            msg = "Not possible to access start_s when initialized from data."
            raise Exception(msg)
        return self._start_position["s"]

    @property
    def start_t(self) -> float:
        if self._initialized_from_data:
            msg = "Not possible to access start_t when initialized from data."
            raise Exception(msg)
        return self._start_position["t"]

    @property
    def start_x(self) -> float:
        if self._initialized_from_data:
            msg = "Not possible to access start_x when initialized from data."
            raise Exception(msg)
        return self._start_position["x"]

    @property
    def start_y(self) -> float:
        if self._initialized_from_data:
            msg = "Not possible to access start_y when initialized from data."
            raise Exception(msg)
        return self._start_position["y"]

    @property
    def start_lat(self) -> float:
        if self._initialized_from_data:
            msg = "Not possible to access start_lat when initialized from data."
            raise Exception(msg)
        return self._start_position["lat"]

    @property
    def start_lon(self) -> float:
        if self._initialized_from_data:
            msg = "Not possible to access start_lon when initialized from data."
            raise Exception(msg)
        return self._start_position["lon"]

    @property
    def z(self) -> float:
        if self._initialized_from_data:
            msg = "Not possible to access z when initialized from data."
            raise Exception(msg)
        return self._start_position["z"]

    @property
    def target_lanelet_id(self) -> int:
        if self._initialized_from_data:
            msg = "Not possible to access target_lanelet_id when initialized from data."
            raise Exception(msg)
        return self._target_position["lanelet_id"]

    @property
    def target_s(self) -> float:
        if self._initialized_from_data:
            msg = "Not possible to access target_s when initialized from data."
            raise Exception(msg)
        return self._target_position["s"]

    @property
    def target_t(self) -> float:
        if self._initialized_from_data:
            msg = "Not possible to access target_t when initialized from data."
            raise Exception(msg)
        return self._target_position["t"]

    @property
    def target_x(self) -> float:
        if self._initialized_from_data:
            msg = "Not possible to access target_x when initialized from data."
            raise Exception(msg)
        return self._target_position["x"]

    @property
    def target_y(self) -> float:
        if self._initialized_from_data:
            msg = "Not possible to access target_y when initialized from data."
            raise Exception(msg)
        return self._target_position["y"]

    @property
    def target_lat(self) -> float:
        if self._initialized_from_data:
            msg = "Not possible to access target_lat when initialized from data."
            raise Exception(msg)
        return self._target_position["lat"]

    @property
    def target_lon(self) -> float:
        if self._initialized_from_data:
            msg = "Not possible to access target_lon when initialized from data."
            raise Exception(msg)
        return self._target_position["lon"]

    @property
    def v0(self) -> float:
        return self._v0

    @property
    def a0(self) -> float:
        return self._a0

    @property
    def inverse_driving_direction(self) -> bool:
        return self._inverse_driving_direction

    @property
    def vehicle_type_name(self) -> str:
        return self._vehicle_type_name

    @property
    def length(self) -> float:
        return self._length

    @property
    def width(self) -> float:
        return self._width

    @property
    def depends_on_ego(self) -> bool:
        return self._depends_on_ego

    @property
    def duration(self) -> float | None:
        return self._duration

    @property
    def x(self) -> np.ndarray:
        self._check_is_compiled()
        return self._x

    @property
    def y(self) -> np.ndarray:
        self._check_is_compiled()
        return self._y

    @property
    def heading(self) -> np.ndarray:
        self._check_is_compiled()
        return self._heading

    @property
    def v(self) -> np.ndarray:
        self._check_is_compiled()
        return self._v

    @property
    def a(self) -> np.ndarray:
        self._check_is_compiled()
        return self._a

    @property
    def initialized_from_data(self) -> bool:
        return self._initialized_from_data

    def _check_is_compiled(self) -> None:
        if not self._compiled:
            msg = "Please call .compile() before accessing this data."
            raise Exception(msg)

    def set_data(
        self, x: np.array, y: np.array, heading: np.array, v: np.array, a: np.array
    ) -> None:
        if not self._initialized_from_data:
            msg = "Cannot set data unless the vehicle is initialized with 'from_data=True'."
            raise Exception(msg)

        self._x = x
        self._y = y
        self._heading = heading
        self._v = v
        self._a = a

        self._compiled = True

    @property
    def compiled(self) -> bool:
        return self._compiled

    def compile(  # noqa: PLR0912
        self,
        road: Road,
        duration: float,
        dt: float,
    ) -> None:
        """
        Create absolute cartesian coordinates etc
        """

        road.resolve_position(self._start_position)
        road.resolve_position(self._target_position)

        full_n_steps = self._n_steps_from_duration_dt(duration, dt)
        active_duration = self._duration if self._duration is not None else duration

        # -- Build time-series arrays --
        n_steps = self._n_steps_from_duration_dt(active_duration, dt)
        if n_steps <= 0:
            pad_len = max(0, full_n_steps)
            pad = np.full((pad_len,), np.nan)
            self._a = pad.copy()
            self._v = pad.copy()
            self._x = pad.copy()
            self._y = pad.copy()
            self._heading = pad.copy()
            self._compiled = True
            return

        # a
        acceleration_vector = self._get_full_acceleration_vector(n_steps, dt)
        speed_change_vector = acceleration_vector * dt

        # v
        speed_vector = self._v0 * np.ones_like(speed_change_vector)
        speed_vector[1:] = speed_vector[1:] + np.cumsum(speed_change_vector)[:-1]
        speed_vector[speed_vector < 1e-2] = 0.0

        # s
        driving_direction = -1 if self._inverse_driving_direction else 1
        lon_position_change_vector = driving_direction * speed_vector * dt
        lon_position_vector = self._start_position["s"] * np.ones_like(
            speed_change_vector
        )
        lon_position_vector[1:] = (
            lon_position_vector[1:] + np.cumsum(lon_position_change_vector)[:-1]
        )

        # t
        lat_offset_vector = self._start_position["t"] * np.ones((n_steps,))

        # -- Determine route --
        source_lanelet = road.lanelet_map.laneletLayer[
            self._start_position["lanelet_id"]
        ]
        if self._target_position["lanelet_id"] is not None:
            target_lanelet = road.lanelet_map.laneletLayer[
                self._target_position["lanelet_id"]
            ]
        else:
            target_lanelet = source_lanelet

        # Establish routing_graph, route without lane changes, get shortest path and transform it into a lanelet_sequence
        routing_graph = lanelet2.routing.RoutingGraph(
            road.lanelet_map, self._traffic_rules
        )
        route = routing_graph.getRoute(source_lanelet, target_lanelet)

        if route is None:
            msg = f"Vehicle {self._vehicle_id}: No valid route found. Please check the route (start_lanelet_id: {self._start_position['lanelet_id']} -> target_lanelet_id: {self._target_position['lanelet_id']})."
            raise ValueError(msg)

        # FOR DEBUGGIN MAPS
        # projector = lanelet2.projection.UtmProjector(lanelet2.io.Origin(50.9098472225444, 6.22742895630397))  # noqa: ERA001
        # lanelet2.io.write("route.osm", route.laneletSubmap().laneletMap(), projector)  # noqa: ERA001
        # lanelet2.io.write("routing_graph.osm", routing_graph.getDebugLaneletMap(0), projector)  # noqa: ERA001
        # FOR DEBUGGIN MAPS

        # -- Handle lane change --
        if self._lc_direction in (-1, 1):
            # Make sure that the requested movement is possible on the route
            # Change the lat offset values according to requested lane change

            # -- Section up to lane change --
            # Find shortest path from source_lanelet to target_lanelet in the route (including lane changes)
            lanelet_path = route.shortestPath()
            # Get initial part of the route until the first lane change
            lanelet_sequence = lanelet_path.getRemainingLane(source_lanelet)

            # Find lanelet at lc position
            lc_start_step = int(self._lc_delay / dt)
            lc_start_s = lon_position_vector[lc_start_step]

            if lc_start_s > lanelet2.geometry.length(lanelet_sequence.centerline):
                msg = f"Vehicle {self._vehicle_id}: Path up to lane change position is not long enough. Please check the route (lanelet_id: {self._start_position['lanelet_id']} -> target_lanelet_id: {self._target_position['lanelet_id']})."
                raise ValueError(msg)

            # -- Lane change section --

            # Find lanelet at lc position
            lc_start_x, lc_start_y = Road.from_frenet_to_cart(
                lanelet_sequence.centerline, lc_start_s, 0
            )
            lc_source_llt_id = Road.find_lanelet_id_by_position_on_lanelet_map(
                road.lanelet_map, lc_start_x, lc_start_y
            )
            if lc_source_llt_id is None:
                msg = f"Vehicle {self._vehicle_id}: Lane change after {self._lc_delay}s at s={lc_start_s}: No valid start lanelet at this position. Maybe there are more than one?"
                raise ValueError(msg)
            lc_source_lanelet = road.lanelet_map.laneletLayer[lc_source_llt_id]

            # Check that there is a valid neighbour for the lane change
            if self._lc_direction == -1:
                lc_target_lanelet = routing_graph.right(lc_source_lanelet)

            elif self._lc_direction == 1:
                lc_target_lanelet = routing_graph.left(lc_source_lanelet)

            if lc_target_lanelet is None:
                direction = "left" if self._lc_direction == 1 else "right"
                msg = f"Vehicle {self._vehicle_id}: Lane change after {self._lc_delay}s at s={lc_start_s} from lanelet {lc_source_llt_id} to the {direction} not possible. No valid neighbour lanelet. (x={lc_start_x}, y={lc_start_y})"
                raise ValueError(msg)

            # Perform the lane change
            if self._lc_type == "polynomial":
                lc_traj = self._generate_lc_trajectory(
                    road.lanelet_map,
                    lon_position_vector[lc_start_step],
                    speed_vector[lc_start_step],
                    lc_source_lanelet.id,
                    lc_target_lanelet.id,
                    self._lc_duration,
                    dt,
                )

                n_lc_steps = lc_traj.shape[0]

                lat_offset_vector[lc_start_step : lc_start_step + n_lc_steps] = lc_traj[
                    :, 1
                ]

            elif self._lc_type == "vy":
                lat_offset_change_vector = (
                    self._lc_vy
                    * dt
                    * np.ones_like(lat_offset_vector)
                    * self._lc_direction
                )
                lat_offset_vector[1:] = (
                    lat_offset_vector[1:] + np.cumsum(lat_offset_change_vector)[:-1]
                )

                # End of LC
                # Find max possible t positon
                x_lc1, y_lc1 = Road.from_frenet_to_cart(
                    lc_source_lanelet.centerline, self._start_position["s"], 0
                )
                _, t_lc1 = Road.from_cart_to_frenet(
                    lc_target_lanelet.centerline, x_lc1, y_lc1
                )
                max_t = t_lc1  # - self.width / 2

                if self._lc_direction == 1:
                    lat_offset_vector[lat_offset_vector > max_t] = max_t
                else:
                    lat_offset_vector[lat_offset_vector < max_t] = max_t

            # -- Section after lane change --

            # Check whether the rest of the route is long enough

            # Find x, y position after lane change
            lc_end_idx = lc_start_step + n_lc_steps - 1
            lc_end_s = lon_position_vector[lc_end_idx]
            lc_end_t = lat_offset_vector[lc_end_idx]
            lc_end_x, lc_end_y = Road.from_frenet_to_cart(
                lc_source_lanelet.centerline, lc_end_s, lc_end_t
            )

            # Find s, t position w.r.t. the lc_target_lanelet
            lc_end_s_in_lc_target_lanelet, lc_end_t_in_lc_target_lanelet = (
                Road.from_cart_to_frenet(
                    lc_target_lanelet.centerline, lc_end_x, lc_end_y
                )
            )

            # Update the remainder of the lon_position_vector and lat_offset_vector to be relative to the new lanelet
            # There may be a s-offset between the lanelets
            lon_position_vector[lc_start_step + n_lc_steps :] = lon_position_vector[
                lc_start_step + n_lc_steps :
            ] + (lc_end_s_in_lc_target_lanelet - lc_end_s)
            # Assuming that the vehicle is staying a the offset t from the end of the lane change
            lat_offset_vector[lc_start_step + n_lc_steps :] = (
                lc_end_t_in_lc_target_lanelet
            )

            # Get remaining route until the target_lanelet
            route_after_lc = routing_graph.getRoute(lc_target_lanelet, target_lanelet)
            shortest_path_after_lc = route_after_lc.shortestPath()
            # Only keep non-lc path
            lanelet_sequence_after_lc = shortest_path_after_lc.getRemainingLane(
                lc_target_lanelet
            )

            remaining_path_length = (
                lanelet2.geometry.length(lanelet_sequence_after_lc.centerline)
                - lc_end_s_in_lc_target_lanelet
            )

            remaining_driven_dist = lon_position_vector[-1] - lc_end_s
            if remaining_driven_dist > remaining_path_length:
                msg = f"Vehicle {self._vehicle_id} is reaching the end of the given path. Please check the route (lanelet_id: {self._start_position['lanelet_id']} -> target_lanelet_id: {self._target_position['lanelet_id']})."
                raise ValueError(msg)

            # -- Compute global x, y for the whole lane change trajectory --

            x = np.zeros_like(lon_position_vector)
            y = np.zeros_like(lat_offset_vector)

            # Until end of lane change
            x[: lc_end_idx + 1], y[: lc_end_idx + 1] = Road.from_frenet_to_cart(
                lanelet_sequence.centerline,
                lon_position_vector[: lc_end_idx + 1],
                lat_offset_vector[: lc_end_idx + 1],
            )
            # After lane change
            x[lc_end_idx + 1 :], y[lc_end_idx + 1 :] = Road.from_frenet_to_cart(
                lanelet_sequence_after_lc.centerline,
                lon_position_vector[lc_end_idx + 1 :],
                lat_offset_vector[lc_end_idx + 1 :],
            )

        else:
            # -- No lane change --

            # Find shortest path from source_lanelet to target_lanelet in the route (including lane changes)
            lanelet_path = route.shortestPath()
            # Get initial part of the route until the first lane change
            lanelet_sequence = lanelet_path.getRemainingLane(source_lanelet)

            # Check that the lanelet_sequence is long enough
            total_driven_dist = lon_position_vector[-1]
            if total_driven_dist > lanelet2.geometry.length(
                lanelet_sequence.centerline
            ):
                msg = f"Vehicle {self._vehicle_id} is reaching the end of the given path. Please check the route (lanelet_id: {self._start_position['lanelet_id']} -> target_lanelet_id: {self._target_position['lanelet_id']})."
                raise ValueError(msg)

            # Positions to cartesian
            x, y = Road.from_frenet_to_cart(
                lanelet_sequence.centerline, lon_position_vector, lat_offset_vector
            )

        # Heading, Assumption: Heading does not change in last time step
        if self._lc_type == "vy":
            if driving_direction == 1:
                heading = np.zeros_like(acceleration_vector)
            elif driving_direction == -1:
                heading = np.pi * np.ones_like(acceleration_vector)
        else:
            # Get initial heading guess from the road
            x_diff0, y_diff0 = Road.from_frenet_to_cart(
                lanelet_sequence.centerline,
                np.array([lon_position_vector[0], lon_position_vector[0] + 0.5]),
                np.zeros((2,)),
            )
            initial_heading_guess = np.arctan2(np.diff(y_diff0), np.diff(x_diff0))
            # Handle standstills (heading remains constant)
            standstill = np.isclose(
                np.diff(x, append=x[-1]) + np.diff(y, append=y[-1]), 0, atol=1e-3
            )
            if np.all(standstill):
                heading = initial_heading_guess * np.ones_like(x)
            else:
                heading = np.arctan2(np.diff(y, append=y[-1]), np.diff(x, append=x[-1]))

                standstill_start_idcs = np.where(
                    np.diff(standstill.astype(int), prepend=0) == 1
                )[0]
                standstill_end_idcs = np.where(
                    np.diff(standstill.astype(int), append=0) == -1
                )[0]
                for start_idx, end_idx in zip(
                    standstill_start_idcs, standstill_end_idcs
                ):
                    if start_idx == 0:
                        value = initial_heading_guess
                    else:
                        value = heading[start_idx - 1]
                    heading[start_idx : end_idx + 1] = value

        self._a = acceleration_vector
        self._v = speed_vector
        self._x = x
        self._y = y
        self._heading = heading

        if n_steps < full_n_steps:
            pad_len = full_n_steps - n_steps
            pad = np.full((pad_len,), np.nan)
            self._a = np.concatenate((self._a, pad))
            self._v = np.concatenate((self._v, pad))
            self._x = np.concatenate((self._x, pad))
            self._y = np.concatenate((self._y, pad))
            self._heading = np.concatenate((self._heading, pad))

        self._compiled = True

    def get_boundary_rect(self) -> tuple[float, float, float, float]:
        valid_mask = ~np.isnan(self._x) & ~np.isnan(self._y)
        if not np.any(valid_mask):
            return (np.nan, np.nan, np.nan, np.nan)
        return (
            np.nanmin(self._x[valid_mask]),
            np.nanmin(self._y[valid_mask]),
            np.nanmax(self._x[valid_mask]),
            np.nanmax(self._y[valid_mask]),
        )

    @staticmethod
    def _n_steps_from_duration_dt(duration: float, dt: float) -> int:
        n_steps = int(np.floor(duration / dt))
        return n_steps

    @staticmethod
    def _elapsed_time_from_n_steps_dt(n_steps: float, dt: float) -> float:
        elapsed_time = np.arange(n_steps, dt)
        return elapsed_time

    def _get_full_acceleration_vector(self, n_steps: int, dt: float) -> np.ndarray:
        if self._a_profile == "constant":
            acceleration_vector = self._a0 * np.ones((n_steps,))

            # Delay
            delay_steps = int(self._a_delay / dt)
            acceleration_vector[:delay_steps] = 0.0

        return acceleration_vector

    def _generate_lc_trajectory(
        self,
        lanelet_map: LaneletMap,
        s_lc0: float,
        v0: float,
        llt_id_lc0: int,
        llt_id_lc1: int,
        lc_duration: float,
        dt: float,
    ) -> np.ndarray:
        """
        Return LC trajectory in frenet frame of llt_lc0 (before lc)
        TODO: Adapt for lower speeds (see sad-rl for solution)
        """
        # !! COPIED FROM: https://gitlab.ika.rwth-aachen.de/lva/pilots/-/blob/main/pilots/highway_pilot/highway_pilot.py?ref_type=heads#L591
        # CHANGED!

        llt_lc0 = lanelet_map.laneletLayer[llt_id_lc0]
        llt_lc1 = lanelet_map.laneletLayer[llt_id_lc1]

        # Keep velocity, constant acceleration over lc_duration
        a = 0
        v1 = v0 + a * lc_duration

        # s position after lc
        s_lc1 = s_lc0 + ((v0 + v1) / 2) * lc_duration

        # Lateral positions in ref_frame
        x_lc0, y_lc0 = Road.from_frenet_to_cart(llt_lc0.centerline, s_lc0, 0)
        x_lc1, y_lc1 = Road.from_frenet_to_cart(llt_lc1.centerline, s_lc1, 0)

        _, t_lc0 = Road.from_cart_to_frenet(llt_lc0.centerline, x_lc0, y_lc0)
        _, t_lc1 = Road.from_cart_to_frenet(llt_lc0.centerline, x_lc1, y_lc1)

        # 5th order polynomial
        d = lc_duration
        p = np.array(
            [
                [0**5, 0**4, 0**3, 0**2, 0**1, 1],  # lateral start position
                [d**5, d**4, d**3, d**2, d**1, 1],  # lateral end positioin
                [0, 0, 0, 0, 1, 0],  # lateral start velocity
                [5 * d**4, 4 * d**3, 3 * d**2, 2 * d, 1, 0],  # lateral end velocity
                [0, 0, 0, 2, 0, 0],  # lateral start acceleration
                [20 * d**3, 12 * d**2, 6 * d, 2, 0, 0],
            ]
        )  # lateral end acceleration
        pv = np.linalg.solve(p, [t_lc0, t_lc1, 0.0, 0.0, 0, 0])

        trajectory_pts = []
        time_steps = np.arange(0, lc_duration + dt, dt)

        for t in time_steps:
            delta_s = s_lc0 + v0 * t + 0.5 * a * t * t
            delta_t = (
                pv[0] * t**5
                + pv[1] * t**4
                + pv[2] * t**3
                + pv[3] * t**2
                + pv[4] * t**1
                + pv[5]
            )
            trajectory_pts.append([delta_s, delta_t])

        trajectory_pts = np.array(trajectory_pts)

        return trajectory_pts

    def _plot_in_ax(
        self,
        ax: plt.Axes,
        timestep: int | None = None,
        include_trace: bool = False,
        color: str | None = None,
    ) -> None:
        if timestep and not (0 <= timestep < self._x.shape[0]):
            msg = "timestep is out of possible range."
            raise ValueError(msg)

        if color is None:
            color = "r"

        if timestep is None:
            if np.isnan(self._x[0]) or np.isnan(self._y[0]):
                return
            x = self._x[0]
            y = self._y[0]
            heading = self._heading[0]
        else:
            if np.isnan(self._x[timestep]) or np.isnan(self._y[timestep]):
                return
            x = self._x[timestep]
            y = self._y[timestep]
            heading = self._heading[timestep]

        rect_patch = Rectangle(
            (x - self.length / 2, y - self.width / 2),
            self.length,
            self.width,
            angle=np.rad2deg(heading),
            rotation_point="center",
            alpha=0.8,
            zorder=30,
            facecolor=color,
            label=f"Vehicle {self._vehicle_id}",
        )

        ax.add_patch(rect_patch)

        # Plot trajectory
        if include_trace:
            if timestep is None:
                ax.plot(
                    self._x, self._y, "-", color=rect_patch.get_facecolor(), zorder=31
                )
            else:
                ax.plot(
                    self._x[timestep:],
                    self._y[timestep:],
                    "-",
                    color=rect_patch.get_facecolor(),
                    zorder=31,
                )
        elif timestep is None:
            ax.plot(self._x, self._y, "x", color=rect_patch.get_facecolor(), zorder=31)

    def _format_ax(self, ax: plt.Axes, *args, **kwargs) -> None:  # noqa: ARG002
        margin = 5
        valid_mask = ~np.isnan(self._x) & ~np.isnan(self._y)
        if not np.any(valid_mask):
            return
        ax.set_xlim(
            np.nanmin(self._x[valid_mask]) - margin,
            np.nanmax(self._x[valid_mask]) + margin,
        )
        ax.set_ylim(
            np.nanmin(self._y[valid_mask]) - margin,
            np.nanmax(self._y[valid_mask]) + margin,
        )

        ax.set_aspect("equal")

        ax.set(title="Vehicle", xlabel="X position in m", ylabel="Y position in m")



from pathlib import Path

from simple_scenario import Scenario, EgoConfiguration, Vehicle
from simple_scenario.road import SyntheticRoad, StraightSegment


class TestOncomingTraffic:
    RESULT_DIR = Path(__file__).parent / "results" / "test_oncoming_traffic"
    RESULT_DIR.mkdir(exist_ok=True)

    def test_scenario_gen(self):
        result_dir = self.RESULT_DIR / "test_scenario_gen"
        result_dir.mkdir(exist_ok=True)

        road = SyntheticRoad(2, 3.75, [StraightSegment(500)])

        ego = EgoConfiguration(
            start_lanelet_id=1000,
            start_s=50,
            start_t=0,
            v0=100 / 3.6,
            target_s=400,
            target_t=0,
        )

        overtaking_vehicle = Vehicle(
            vehicle_id=0,
            start_lanelet_id=1001,
            start_s=400,
            start_t=0,
            v0=100 / 3.6,
            lc_delay=3,
            lc_direction=-1,
            inverse_driving_direction=True,
        )
        slow_vehicle = Vehicle(
            vehicle_id=1,
            start_lanelet_id=1001,
            start_s=330,
            start_t=0.5,
            v0=60 / 3.6,
            inverse_driving_direction=True,
        )

        scenario = Scenario(
            "oncoming_traffic", road, ego, [overtaking_vehicle, slow_vehicle], 13
        )

        scenario.render_gif(result_dir)

        for vehicle in scenario.vehicles:
            if vehicle.inverse_driving_direction:
                assert vehicle.x[-1] < vehicle.x[0], (
                    f"Vehicle {vehicle.id} should drive in inverse driving direction, but it is not."
                )

        print("TEST PASSED")


if __name__ == "__main__":
    tester = TestOncomingTraffic()
    tester.test_scenario_gen()



from .arc_segment import ArcSegment
from .clothoid_segment import ClothoidSegment
from .mapped_road import MappedRoad
from .polyline_segment import PolylineSegment
from .road_segment import RoadSegment
from .straight_segment import StraightSegment
from .synthetic_road import SyntheticRoad
